import com.mindfusion.diagramming.*;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.sql.*;
import java.util.*;

class Crud {

    static void save(Diagram diagram){

        String diagramQuery = "INSERT INTO diagrams (id, name) values (?,?)";

        String name = diagram.getTag().toString();

        Connection connection = null;
        PreparedStatement statement = null;
        try{
            connection = DriverManager.getConnection(Credentials.path, Credentials.user, Credentials.password);

            if (!diagramKeyMap.containsValue(diagram.getTag().toString())){
                diagramKeyMap.put(diagramKeyMap.isEmpty() ? 1 : diagramKeyMap.lastKey() + 1, name);
                diagramValueMap.put(name, diagramKeyMap.lastKey());
                statement = connection.prepareStatement(diagramQuery);
                statement.setInt(1, diagramKeyMap.lastKey());
                statement.setString(2, name);
                statement.execute();
            }

            DiagramNodeList diagramNodes = diagram.getNodes();
            for (DiagramNode node : diagramNodes) {
                if (node.getTag() == null && !savedNodesMap.containsKey((String) node.getId()))
                    insert(node, name);
                else if (updatedNodeMap.containsKey((String) node.getId()))
                    update(node, name);
            }
            DiagramLinkList diagramLinks = diagram.getLinks();
            for (DiagramLink link : diagramLinks){
                if (link.getTag() == null && !savedLinksMap.containsKey((String)link.getId()))
                    insert(link, name);
                else if (updatedLinkMap.containsKey((String)link.getId()))
                    update(link, name);
            }

            for (DiagramNode node : deletedNodeMap.values())
                delete(node, name);
            for (DiagramLink link : deletedLinkMap.values())
                delete(link, name);

            updatedNodeMap.clear();
            deletedNodeMap.clear();
            updatedLinkMap.clear();
            deletedLinkMap.clear();

        }catch (SQLException e){
            e.printStackTrace();
        }finally {

            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    e.printStackTrace();
                }
            }
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    e.printStackTrace();
                }
            }
        }

    }


    private static void insert(DiagramItem item, String diagramName) throws SQLException {

        String nodesQuery = "INSERT INTO nodes(id, diagramId, text, shape, x, y, width, height) VALUES (?,?,?,?,?,?,?,?)";
        String linksQuery = "INSERT INTO links(id, diagramId, originId, destinationId, shape, text) VALUES (?,?,?,?,?,?)";
        String controlsQuery = "INSERT INTO controlpoints (id, linkId, pointIndex, x, y) VALUES (?,?,?,?,?)";

        Connection connection = DriverManager.getConnection(Credentials.path, Credentials.user, Credentials.password);
        PreparedStatement statement;
        PreparedStatement cStatement;
        switch (item.getClass().getSimpleName()) {
            case "ShapeNode":
                statement = connection.prepareStatement(nodesQuery);

                ShapeNode shapeNode = (ShapeNode) item;
                statement.setString(1, shapeNode.getId().toString());
                statement.setInt(2, diagramValueMap.get(diagramName));
                statement.setString(3, shapeNode.getText());
                statement.setString(4, shapeNode.getShape().getId());
                statement.setDouble(5, shapeNode.getBounds().getX());
                statement.setDouble(6, shapeNode.getBounds().getY());
                statement.setDouble(7, shapeNode.getBounds().getWidth());
                statement.setDouble(8, shapeNode.getBounds().getHeight());
                statement.execute();

                savedNodesMap.put(shapeNode.getId().toString(), shapeNode);
                shapeNode.setTag(diagramValueMap.get(diagramName));

                connection.close();
                statement.close();
                break;
            case "DiagramLink":
                statement = connection.prepareStatement(linksQuery);
                cStatement = connection.prepareStatement(controlsQuery);

                DiagramLink link = (DiagramLink) item;
                statement.setString(1, link.getId().toString());
                statement.setInt(2, diagramValueMap.get(diagramName));
                statement.setString(3, link.getOrigin().getId().toString());
                statement.setString(4, link.getDestination().getId().toString());
                statement.setString(5, link.getShape().name());
                statement.setString(6, link.getText());
                statement.execute();
                for (int i = 0; i < link.getControlPoints().size(); i++) {
                    cStatement.setInt(1, diagramValueMap.get(diagramName));
                    cStatement.setString(2, link.getId().toString());
                    cStatement.setInt(3, i);
                    Point2D.Float points = link.getControlPoints().get(i);
                    cStatement.setDouble(4, points.getX());
                    cStatement.setDouble(5, points.getY());
                    cStatement.execute();
                }
                savedLinksMap.put(link.getId().toString(), link);
                link.setTag(diagramValueMap.get(diagramName));

                connection.close();
                statement.close();
                cStatement.close();
                break;
        }
    }
    private static void update(DiagramItem item, String diagramName) throws SQLException{
        String updateNodes = "UPDATE nodes SET  text = ?, shape = ?, x = ?, y = ?, width = ?, height = ? WHERE  id = ? AND  diagramId = ?";
        String updateControls = "UPDATE controlpoints SET  x = ? , y = ? WHERE  id = ? AND linkId = ? AND pointIndex = ?";
        String updateLinks = "UPDATE links SET  originId = ?, destinationId = ?, shape = ?, text = ? WHERE id = ? AND diagramId = ?";

        Connection connection = DriverManager.getConnection(Credentials.path, Credentials.user, Credentials.password);
        PreparedStatement statement;
        PreparedStatement cStatement = null;

        switch (item.getClass().getSimpleName()){
            case "ShapeNode":
                statement = connection.prepareStatement(updateNodes);

                ShapeNode node = (ShapeNode) item;
                Rectangle2D.Float fl = node.getBounds();
                statement.setString(1, node.getText());
                statement.setString(2, node.getShape().getId());
                statement.setDouble(3, fl.getX());
                statement.setDouble(4, fl.getY());
                statement.setDouble(5, fl.getWidth());
                statement.setDouble(6, fl.getHeight());
                statement.setString(7, (String)node.getId());
                statement.setInt(8, diagramValueMap.get(diagramName));
                statement.execute();
                DiagramLinkList linkList = node.getIncomingLinks();
                if (!linkList.isEmpty())
                    cStatement = connection.prepareStatement(updateControls);
                for (DiagramLink link : linkList){

                    for (int i=0; i<link.getControlPoints().size(); i++){
                        Point2D.Float points = link.getControlPoints().get(i);
                        cStatement.setDouble(1,  points.getX());
                        cStatement.setDouble(2,  points.getY());
                        cStatement.setInt(3, diagramValueMap.get(diagramName));
                        cStatement.setString(4, link.getId().toString());
                        cStatement.setInt(5, i);
                        cStatement.execute();
                    }
                }
                linkList = node.getOutgoingLinks();
                if (!linkList.isEmpty())
                    cStatement = connection.prepareStatement(updateControls);
                for (DiagramLink link : linkList){
                    for (int i = 0; i<link.getControlPoints().size(); i++){
                        Point2D.Float points = link.getControlPoints().get(i);
                        cStatement.setDouble(1,  points.getX());
                        cStatement.setDouble(2,  points.getY());
                        cStatement.setInt(3, diagramValueMap.get(diagramName));
                        cStatement.setString(4, link.getId().toString());
                        cStatement.setInt(5, i);
                        cStatement.execute();
                    }
                }
                connection.close();
                statement.close();
                if (cStatement != null)
                    cStatement.close();
                break;
            case "DiagramLink":
                statement = connection.prepareStatement(updateLinks);
                cStatement = connection.prepareStatement(updateControls);

                DiagramLink link = (DiagramLink) item;
                statement.setString(1, link.getOrigin().getId().toString());
                statement.setString(2, link.getDestination().getId().toString());
                statement.setString(3, link.getShape().name());
                statement.setString(4, link.getText());
                statement.setString(5, (String)link.getId());
                statement.setInt(6, diagramValueMap.get(diagramName));
                statement.execute();

                for (int i = 0; i<link.getControlPoints().size() ; i++){
                    Point2D.Float points = link.getControlPoints().get(i);
                    cStatement.setDouble(1, points.getX());
                    cStatement.setDouble(2, points.getY());
                    cStatement.setInt(3, diagramValueMap.get(diagramName));
                    cStatement.setString(4, (String)link.getId());
                    cStatement.setInt(5, i);
                    cStatement.execute();
                }
                connection.close();
                statement.close();
                cStatement.close();
                break;
        }

    }

    private static void delete(DiagramItem item, String diagramName) throws SQLException {

        String deleteNodes = "DELETE FROM nodes WHERE id = ? AND diagramId = ?";
        String deleteLinks = "DELETE links, controlpoints FROM links " +
                "INNER JOIN controlpoints ON links.id = controlpoints.linkId " +
                "WHERE links.diagramId = ? AND controlpoints.id = ? AND links.id = ? AND controlpoints.linkId = ?";

        Connection connection = DriverManager.getConnection(Credentials.path, Credentials.user, Credentials.password);
        PreparedStatement statement;

        switch (item.getClass().getSimpleName()){
            case "ShapeNode":
                ShapeNode node = (ShapeNode) item;

                statement = connection.prepareStatement(deleteNodes);
                statement.setString(1, (String)node.getId());
                statement.setInt(2, diagramValueMap.get(diagramName));
                statement.execute();

                savedNodesMap.remove((String)node.getId());

                connection.close();
                statement.close();
                break;
            case "DiagramLink":
                DiagramLink link = (DiagramLink) item;

                statement = connection.prepareStatement(deleteLinks);
                statement.setInt(1, diagramValueMap.get(diagramName));
                statement.setInt(2, diagramValueMap.get(diagramName));
                statement.setString(3, (String)link.getId());
                statement.setString(4, (String)link.getId());
                statement.execute();

                savedLinksMap.remove((String)link.getId());

                connection.close();
                statement.close();
                break;
        }
    }


    static boolean load(int id, String name, Diagram diagram){

        boolean bo = true;
        savedNodesMap.clear();
        updatedNodeMap.clear();
        deletedNodeMap.clear();
        savedLinksMap.clear();
        updatedLinkMap.clear();
        deletedLinkMap.clear();
        // load nodes
        String nodesQuery = "SELECT nodes.x, nodes.y, nodes.width, nodes.height, nodes.id, nodes.shape, nodes.text " +
                "FROM diagrams " +
                "LEFT JOIN nodes " +
                "ON diagrams.id = nodes.diagramId " +
                "WHERE diagrams.id =  ? AND diagrams.name = ?";
        String linkQuery = "SELECT links.id, links.originId, links.destinationId, links.shape, links.text, " +
                "GROUP_CONCAT(pointIndex), GROUP_CONCAT(x), GROUP_CONCAT(y) " +
                "FROM diagrams " +
                "LEFT JOIN links " +
                "ON diagrams.id = links.diagramId " +
                "LEFT JOIN controlpoints " +
                "ON links.id = controlpoints.linkId " +
                "WHERE diagrams.id = ? AND links.diagramId = ? AND controlpoints.id = ? AND diagrams.name = ? " +
                "GROUP BY links.id";

        Connection connection = null;
        PreparedStatement statement = null;
        try{
            connection = DriverManager.getConnection(Credentials.path, Credentials.user, Credentials.password);
            statement = connection.prepareStatement(nodesQuery);
            statement.setInt(1, id);
            statement.setString(2, name);
            ResultSet itemset = statement.executeQuery();
            while (itemset.next()){
                ShapeNode node = diagram.getFactory().createShapeNode(
                        itemset.getDouble(1),
                        itemset.getDouble(2),
                        itemset.getDouble(3),
                        itemset.getDouble(4));
                node.setId(itemset.getString(5));
                node.setShape(Shape.fromId(itemset.getString(6)));
                if (itemset.getString(7) != null)
                    node.setText(itemset.getString(7));
                node.setTag(id);
                savedNodesMap.put(node.getId().toString(), node);
            }
            statement = connection.prepareStatement(linkQuery);
            statement.setInt(1, id);
            statement.setInt(2, id);
            statement.setInt(3, id);
            statement.setString(4, name);
            itemset = statement.executeQuery();
            while (itemset.next()){
                DiagramLink diagramLink = new DiagramLink(diagram);
                diagramLink.setId(itemset.getString(1));
                diagramLink.setOrigin(savedNodesMap.get(itemset.getString(2)));
                diagramLink.setDestination(savedNodesMap.get(itemset.getString(3)));
                diagramLink.setShape(LinkShape.valueOf(itemset.getString(4)));
                diagramLink.setHeadShape(Shape.fromId("Arrow"));
                if (itemset.getString(5) != null)
                    diagramLink.setText(itemset.getString(5));
                String [] indexes = itemset.getString(6).split(",");
                String [] xPoints = itemset.getString(7).split(",");
                String [] yPoints = itemset.getString(8).split(",");
                diagramLink.getControlPoints().clear();
                for (int i =0; i<indexes.length; i++){
                    Point2D.Float point2D = new Point2D.Float();
                    point2D.setLocation(Double.valueOf(xPoints[i]), Double.valueOf(yPoints[i]));
                    diagramLink.getControlPoints().add(i, point2D);
                }
                diagramLink.setTag(id);
                diagram.getLinks().add(diagramLink);
                savedLinksMap.put(diagramLink.getId().toString(), diagramLink);

            }

        }catch (SQLException e){
            System.err.println(e.getMessage());
            bo = false;
        }finally {
            if (connection != null){
                try {
                    connection.close();
                } catch (SQLException e) {
                    System.err.println(e.getMessage());
                    bo = false;
                }
            }
            if (statement != null){
                try {
                    statement.close();
                } catch (SQLException e) {
                    System.err.println(e.getMessage());
                    bo = false;
                }
            }
        }

        return bo;
    }

    static Vector<String> getDiagrams() {

        String namesSQL = "SELECT id, name from diagrams";
        Connection connection = null;
        PreparedStatement statement = null;
        diagramKeyMap.clear();
        diagramValueMap.clear();
        try {
            connection = DriverManager.getConnection(Credentials.path, Credentials.user, Credentials.password);
            statement = connection.prepareStatement(namesSQL);
            ResultSet resultSet = statement.executeQuery();
            while (resultSet.next()){
                diagramKeyMap.put(resultSet.getInt(1), resultSet.getString(2));
                diagramValueMap.put(resultSet.getString(2), resultSet.getInt(1));
            }
        } catch (SQLException ignore) {

        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException e) {
                    e.printStackTrace();
                }
            }
            if (statement != null) {
                try {
                    statement.close();
                } catch (SQLException e) {
                    e.printStackTrace();
                }
            }
        }
        return new Vector<>(diagramKeyMap.values());
    }

    static TreeMap<Integer, String> diagramKeyMap = new TreeMap<>();
    static Map<String, Integer> diagramValueMap = new Hashtable<>();

    static Map<String, DiagramLink> savedLinksMap = new Hashtable<>();
    static Map<String, DiagramLink> updatedLinkMap = new Hashtable<>();
    static Map<String, DiagramLink> deletedLinkMap = new Hashtable<>();

    static Map<String, DiagramNode> savedNodesMap = new Hashtable<>();
    static Map<String, DiagramNode> updatedNodeMap = new Hashtable<>();
    static Map<String, DiagramNode> deletedNodeMap = new Hashtable<>();
}
